﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ELTE.Algorithms.BinarySearchTree.Traversal.Model
{
    /// <summary>
    /// Bináris keresőfa elemtípusa.
    /// </summary>
    public class TreeNode
    {
        /// <summary>
        /// Szülő mutató.
        /// </summary>
        public TreeNode Parent;
        /// <summary>
        /// Bal gyerek mutató.
        /// </summary>
        public TreeNode Left;
        /// <summary>
        /// Jobb gyerek mutató.
        /// </summary>
        public TreeNode Right;
        /// <summary>
        /// Kulcs értéke.
        /// </summary>
        public Int32 Key;

        public TreeNode(Int32 key) { Key = key; }
        public TreeNode(Int32 key, TreeNode parent) { Key = key; Parent = parent; }
        public TreeNode(Int32 key, TreeNode parent, TreeNode left, TreeNode right) 
        { 
            Key = key; 
            Parent = parent;
            Left = left;
            Right = right;
        }
    }

    /// <summary>
    /// Bináris keresőfa típusa.
    /// </summary>
    public class BinarySearchTree
    {
        protected TreeNode _Root; // gyökérelem
        protected TreeNode _Current; // aktuális csúcs
        
        /// <summary>
        /// Fa gyökércsúcsának lekrédezése.
        /// </summary>
        public TreeNode Root { get { return _Root; } }
        /// <summary>
        /// Aktuális csúcs lekérdezése.
        /// </summary>
        public TreeNode Current { get { return _Current; } }
        /// <summary>
        /// Fa ürességének lekrédezése.
        /// </summary>
        public Boolean IsEmpty { get { return (_Root == null); } }

        /// <summary>
        /// Új bináris keresőfa létrehozása.
        /// </summary>
        public BinarySearchTree()
        {
            _Root = _Current = null;
        }

        /// <summary>
        /// Beszúrás.
        /// </summary>
        /// <param name="key">Kulcs értéke.</param>
        /// <returns>A beszúrt csúcs mutatója.</returns>
        public TreeNode Insert(Int32 key)
        {
            TreeNode y = null, x = _Root;
            while (x != null && x.Key != key)
            {
                y = x;
                if (key < x.Key)
                    x = x.Left;
                else
                    x = x.Right;
            }
            if (x == null)
            {
                if (y == null)
                    _Root = _Current = new TreeNode(key);
                else
                {
                    if (key < y.Key)
                        y.Left = _Current = new TreeNode(key, y);
                    else
                        y.Right = _Current = new TreeNode(key, y);
                }
                return _Current;
            }
            return null;
        }
        /// <summary>
        /// Minimum elem kiválasztása.
        /// </summary>
        /// <returns>Minimum elem mutatója.</returns>
        public TreeNode Min()
        {
            if (_Root == null)
                return null;
            _Current = _Root;
            while (_Current.Left != null)
                _Current = _Current.Left;
            return _Current;
        }
        /// <summary>
        /// Maximum elem kiválasztása.
        /// </summary>
        /// <returns>Maximum elem mutatója.</returns>
        public TreeNode Max()
        {
            if (_Root == null)
                return null;
            _Current = _Root;
            while (_Current.Right != null)
                _Current = _Current.Right;
            return _Current;
        }
        /// <summary>
        /// Előző elem kiválassztása (ha van).
        /// </summary>
        /// <returns>Előző elem mutatója.</returns>
        public TreeNode Previous()
        {
            if (_Current != null)
            {
                if (_Current.Left != null)
                {
                    _Current = _Current.Left;
                    while (_Current.Right != null)
                        _Current = _Current.Right;
                }
                else
                {
                    TreeNode p = _Current, s = _Current.Parent;
                    while (s != null && s.Left == p)
                    {
                        p = s;
                        s = s.Parent;
                    }
                    if (s != null)
                        _Current = s;
                }
            }
            return _Current;
        }
        /// <summary>
        /// Következő elem kiválassztása (ha van).
        /// </summary>
        /// <returns>Következő elem mutatója.</returns>
        public TreeNode Next()
        {
            if (_Current != null)
            {
                if (_Current.Right != null)
                {
                    _Current = _Current.Right;
                    while (_Current.Left != null)
                        _Current = _Current.Left;
                }
                else
                {
                    TreeNode p = _Current, s = _Current.Parent;
                    while (s != null && s.Right == p)
                    {
                        p = s;
                        s = s.Parent;
                    }
                    if (s != null)
                        _Current = s;
                }
            }
            return _Current;
        }
        /// <summary>
        /// Keresés.
        /// </summary>
        /// <param name="key">Keresett kulcs.</param>
        /// <returns>Keresett elem mutatója.</returns>
        public TreeNode Search(Int32 key)
        {
            _Current = _Root;
            while (_Current != null && _Current.Key != key)
            {
                if (key < _Current.Key)
                    _Current = _Current.Left;
                else
                    _Current = _Current.Right;
            }
            return _Current;
        }
        /// <summary>
        /// Törlés.
        /// </summary>
        /// <param name="key">Törlendő kulcs.</param>
        /// <returns>Kitörölt elem mutatója.</returns>
        public TreeNode Delete(Int32 key)
        { 
            TreeNode p = Search(key);
            TreeNode x = null, y = null;

            if (p != null)
            {
                if (p.Left == null || p.Right == null)
                    y = p;
                else
                    y = Next();
            }
            if (p != null)
            {
                if (y.Left != null)
                    x = y.Left;
                else
                    x = y.Right;
                if (x != null)
                    x.Parent = y.Parent;
                if (y.Parent == null)
                    _Root = x;
                else
                    if (y == y.Parent.Left)
                        y.Parent.Left = x;
                    else
                        y.Parent.Right = x;
                if (y != p)
                    p.Key = y.Key;
            }
            if (x != null)
                _Current = x;
            else
                _Current = y.Parent;
            return y;
        }
        /// <summary>
        /// Fa kiürítése.
        /// </summary>
        public void Clear()
        {
            _Root = _Current = null;
        }

        /// <summary>
        /// Fa kirajzolása.
        /// </summary>
        /// <param name="gr">Rajzobjektum, amivel rajzolunk.</param>
        /// <param name="width">A szélesség, amelyben rajzolunk.</param>
        public void Draw(Graphics gr, Int32 width)
        {
            if (_Root != null) // ha van elem a fában
                DrawNode(_Root, gr, 0, width, 20); // meghívunk egy rekurzív eljárást, amely egy csúcsot ki tud rajzolni
        }
        private void DrawNode(TreeNode node, Graphics gr, Int32 start, Int32 width, Int32 height)
        // egy csúcs és hozzá tartozó élek kirajzolása, megkapja, hogy melyik csúcsot, milyen rajzobjektummal, milyen kezdőpozícióval, 
        // milyen szélességben és milyen magasságban kell kirajzolnia
        { 
            Font font = new Font(FontFamily.GenericSansSerif, 8, FontStyle.Bold); // betűtípus
            Pen bluePen = new Pen(Color.Blue); // kék toll a csúcs és élrajzoláshoz
            Pen redPen = new Pen(Color.Red); // piros toll az aktuálisan kijelölt csúcsnak
            SizeF keySize = gr.MeasureString(node.Key.ToString(), font); // le kell mérnünk, mekkora a csúcs kiírása ahhoz, hogy középre rakhassuk

            if (node.Left != null) // ha van gyerek, akkor azt is ki kell rajzolnunk
            {
                gr.DrawLine(bluePen, start + width / 2, height, start + width / 4, height + 50);
                // él kirajzolása a két csúcs közé
                DrawNode(node.Left, gr, start, width / 2, height + 50); // rekurzívan meghívjuk az eljárást a bal gyerek csúcsra
            }
            if (node.Right != null) // ugyaanez a jobb gyerek csúcsra
            {
                gr.DrawLine(bluePen, start + width / 2, height, start + 3 * width / 4, height + 50);
                DrawNode(node.Right, gr, start + width / 2, width / 2, height + 50);
            }

            gr.FillEllipse(Brushes.White, start + width / 2 - 15, height - 15, 30, 30); // háttér újrarajzolása
            if (node == _Current) // ha az aktuális csúcsot rajzoljuk éppen, akkor pirossal
                gr.DrawEllipse(redPen, start + width / 2 - 15, height - 15, 30, 30);
            else // különben kékkel
                gr.DrawEllipse(bluePen, start + width / 2 - 15, height - 15, 30, 30);
            gr.DrawString(node.Key.ToString(), font, Brushes.Black, start + width / 2 - keySize.Width / 2, height - keySize.Height / 2); // belerajzoljuk a kulcsot
        }
    }
}
